/*
 Adjustable 2A PSU
 Version: 1.0
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 16/10/2017
 Last Modified: 23/02/2018

 */
 
#ifndef cbi
#define cbi(sfr, bit) (_SFR_uint8_t(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_uint8_t(sfr) |= _BV(bit))
#endif

#ifndef boolean
	typedef uint8_t boolean;
#endif
#ifndef bool
	typedef uint8_t bool;
#endif 
#ifndef uint8_t
	typedef uint8_t uint8_t;
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

#define LOW 0
#define HIGH 1
#define false 0
#define true 1

#define EN_0_OHM PD5
#define EN_4_7_OHM PD4
#define EN_10_OHM PD3
#define EN_20_OHM PD2

#define LED_ENABLE PD6
#define SW_ENABLE PD7
#define SW_SELECT PB2

// MCP4151
#define DPOT_CLK PD0
#define DPOT_CS PD1
#define DPOT_DATA PC5

#define DPOT_WIPER0 0 
#define DPOT_WRITEDATA 0

uint8_t dpotValue = 50; // Starting value
uint8_t dpotValueOld = 0;
uint8_t mosfetEnabled = 0;
uint8_t mosfetEnabledOld = 0;

// DAC
#define WRITE_DAC_REG 0  // 00 command bits, 00 power down bits (Not Powered Down)
#define WRITE_DAC_CONFIG_ENABLE_VREF 0x98  // 100 command bits, 11 VRef, 00 power down bits (Not Powered Down), 0 Gain

#define DAC_ADDR 0x61

#define DAC_SCL PB4
#define DAC_SDA PB5

#define I2C_DELAY_USEC 4
#define I2C_READ 1
#define I2C_WRITE 0

// Other Vars
uint16_t adcVoltageOld = 0;
uint8_t inCalibration = false;
uint8_t calibrated = false;
uint8_t waitForLowVoltage = false;

uint8_t calibrationValues[400][2];
uint16_t calibrationCounter = 0;
uint16_t calibrationLimit = 0;

volatile uint8_t outputEnabled = false;
volatile uint8_t encoderWaitforlow = false;
volatile uint8_t encoderTimeout = true;

#define CHANGE_VOLTAGE 1
#define CHANGE_CURRENT 2
uint8_t selectedOption = CHANGE_VOLTAGE;
uint16_t voltageNumber = 000;
uint16_t currentNumber = 000;
volatile uint16_t currentLimit = 900; // ~500mA current limit at start up
uint16_t currentLimitOld = 0;
uint8_t currentLimited = false;
volatile uint8_t settingCurrent = 0; // Used to show the current being set over the current in use
int currentLimitCalc;

// SPI write
void spi_write_msb(uint8_t data) {
	for (int8_t i = 7; i >= 0; i--) {
		if (data & (1<<i)) {
			PORTC |= (1<<DPOT_DATA);
		} else {
			PORTC &= ~(1<<DPOT_DATA);
		}
		PORTD |= (1<<DPOT_CLK);
		PORTD &= ~(1<<DPOT_CLK);
	}
}

// Initialize SCL/SDA pins and set the bus high
void SoftI2cMasterInit(uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	DDRB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SDA_PIN);
	DDRB |= (1<<TWI_SCL_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
}

// De-initialize SCL/SDA pins and set the bus low
void SoftI2cMasterDeInit(uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	DDRB &= ~(1<<TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SCL_PIN);
	DDRB &= ~(1<<TWI_SCL_PIN);
}

// Read a byte from I2C and send Ack if more reads follow else Nak to terminate read
uint8_t SoftI2cMasterRead(uint8_t last, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	uint8_t b = 0;
	// Make sure pull-up enabled
	PORTB |= (1<<TWI_SDA_PIN);
	DDRB &= ~(1<<TWI_SDA_PIN);
	// Read byte
	for (uint8_t i = 0; i < 8; i++) {
		// Don't change this loop unless you verify the change with a scope
		b <<= 1;
		_delay_us(I2C_DELAY_USEC);
		PORTB |= (1<<TWI_SCL_PIN);
		if (bit_is_set(PINB, TWI_SDA_PIN)) b |= 1;
		PORTB &= ~(1<<TWI_SCL_PIN);
	}
	// Send Ack or Nak
	DDRB |= (1<<TWI_SDA_PIN);
	if (last) { 
		PORTB |= (1<<TWI_SDA_PIN); 
	}
	else { 
		PORTB &= ~(1<<TWI_SDA_PIN);
	}  
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB &= ~(1<<TWI_SCL_PIN);
	PORTB &= ~(1<<TWI_SDA_PIN);
	return b;
}

// Write a byte to I2C
uint8_t SoftI2cMasterWrite(uint8_t data, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	// Write byte
	for (uint8_t m = 0x80; m != 0; m >>= 1) {
		// Don't change this loop unless you verify the change with a scope
		if (m & data) { 
			PORTB |= (1<<TWI_SDA_PIN); 
		}
		else { 
			PORTB &= ~(1<<TWI_SDA_PIN); 
		}
		PORTB |= (1<<TWI_SCL_PIN);
		_delay_us(I2C_DELAY_USEC);
		PORTB &= ~(1<<TWI_SCL_PIN);
	}
	// get Ack or Nak
	DDRB &= ~(1<<TWI_SDA_PIN);
	// Enable pullup
	PORTB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
	uint8_t rtn = bit_is_set(PINB, TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SCL_PIN);
	DDRB |= (1<<TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SDA_PIN);
	return rtn == 0;
}

// Issue a start condition
uint8_t SoftI2cMasterStart(uint8_t addressRW, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB &= ~(1<<TWI_SCL_PIN);
	return SoftI2cMasterWrite(addressRW, TWI_SCL_PIN, TWI_SDA_PIN);
}

// Issue a restart condition
uint8_t SoftI2cMasterRestart(uint8_t addressRW, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	PORTB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	return SoftI2cMasterStart(addressRW, TWI_SCL_PIN, TWI_SDA_PIN);
}

// Issue a stop condition
void SoftI2cMasterStop(uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB |= (1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
}

// Write 1 byte, no address
uint8_t soft_i2c_write_byte(uint8_t deviceAddr, uint8_t writeByte, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	// Issue a start condition, send device address and write direction bit
	if (!SoftI2cMasterStart((deviceAddr<<1) | I2C_WRITE, TWI_SCL_PIN, TWI_SDA_PIN)) return false;
	
	// Write the byte
	if (!SoftI2cMasterWrite(writeByte, TWI_SCL_PIN, TWI_SDA_PIN)) return false;

	// Issue a stop condition
	SoftI2cMasterStop(TWI_SCL_PIN, TWI_SDA_PIN);
	
	return true;
}

// Write 2 bytes, no address
uint8_t soft_i2c_write_2_bytes(uint8_t deviceAddr, uint8_t writeByte1, uint8_t writeByte2, uint8_t TWI_SCL_PIN, uint8_t TWI_SDA_PIN) {
	// Issue a start condition, send device address and write direction bit
	if (!SoftI2cMasterStart((deviceAddr<<1) | I2C_WRITE, TWI_SCL_PIN, TWI_SDA_PIN)) return false;
	
	// Write the byte1
	if (!SoftI2cMasterWrite(writeByte1, TWI_SCL_PIN, TWI_SDA_PIN)) return false;
	
	// Write the byte2
	if (!SoftI2cMasterWrite(writeByte2, TWI_SCL_PIN, TWI_SDA_PIN)) return false;

	// Issue a stop condition
	SoftI2cMasterStop(TWI_SCL_PIN, TWI_SDA_PIN);
	
	return true;
}

// Update the voltage dpot / mosfet selections
void update_voltage_dpot_mosfet(void) {
	// If calibrated, update the dpot / mosfet values
	if (calibrated == true) {
		dpotValue = calibrationValues[calibrationCounter][0];
		mosfetEnabled = calibrationValues[calibrationCounter][1];
	}
	
	// Dpot value / mosfet enabled changed?
	if (dpotValue != dpotValueOld || mosfetEnabled != mosfetEnabledOld) {
		PORTD &= ~(1<<DPOT_CS);
		spi_write_msb((DPOT_WIPER0 << 4) | (DPOT_WRITEDATA << 2));// | (dpotValue >> 8));
		spi_write_msb(dpotValue & 0xFF);
		PORTD |= (1<<DPOT_CS);
		
		// More resistance = voltage goes higher
		if (mosfetEnabled == 0) {
			PORTD |= (1<<EN_20_OHM);
			PORTD &= ~((1<<EN_10_OHM) | (1<<EN_4_7_OHM) | (1<<EN_0_OHM));
		}
		else if (mosfetEnabled == 1) {
			PORTD |= (1<<EN_10_OHM);
			PORTD &= ~((1<<EN_20_OHM) | (1<<EN_4_7_OHM) | (1<<EN_0_OHM));
		}
		else if (mosfetEnabled == 2) {
			PORTD |= (1<<EN_4_7_OHM);
			PORTD &= ~((1<<EN_20_OHM) | (1<<EN_10_OHM) | (1<<EN_0_OHM));
		}
		else if (mosfetEnabled == 3) { // 0 ohm
			PORTD |= (1<<EN_0_OHM);
			PORTD &= ~((1<<EN_20_OHM) | (1<<EN_10_OHM) | (1<<EN_4_7_OHM));
		}
		
		dpotValueOld = dpotValue;
		mosfetEnabledOld = mosfetEnabled;
	}
}

// Update the current limit
void update_current_limit(void) {
	// Current limit changed?
	if (currentLimit != currentLimitOld) {
		soft_i2c_write_2_bytes(DAC_ADDR, WRITE_DAC_REG | ((currentLimit & 0xF00) >> 8), (currentLimit & 0xFF), DAC_SCL, DAC_SDA);
		currentLimitOld = currentLimit;
	}
}

void display_screen(void) {
	Adafruit_SSD1306_clearDisplay();
	xPos = 0;
	yPos = 0;
	
	// Voltage
	if ((voltageNumber / 1000) >= 1) { // Over 10V
		displayChar((int) voltageNumber / 1000);
		displayChar((int) (voltageNumber / 100) % 10);
		drawDot(30, 0);
		xPos += 5;
		
		displayChar((int) (voltageNumber / 10) % 10);
		displayChar((int) voltageNumber % 10);
		displayChar(FONT_V);
	}
	else {
		displayChar((int) (voltageNumber / 100) % 10);
		drawDot(15, 0);
		xPos += 5;
		
		displayChar((int) (voltageNumber / 10) % 10);
		displayChar((int) voltageNumber % 10);
		displayChar(0);
		displayChar(FONT_V);
	}
	if (selectedOption == CHANGE_VOLTAGE) {
		xPos += 5;
		displayChar(FONT_ARROW);
	}
	
	
	// Current
	xPos = 0;
	yPos = 22;
	
	// Show setting current
	currentLimitCalc = currentLimit / 1.6;
	if (settingCurrent >= 1 && settingCurrent < 5) {
		displayChar((int) currentLimitCalc / 1000);
		drawDot(15, 22);
		xPos += 5;
		
		displayChar((int) (currentLimitCalc / 100) % 10);
		displayChar((int) (currentLimitCalc / 10) % 10);
		displayChar((int) currentLimitCalc % 10);
		displayChar(FONT_A);
		
		settingCurrent++;
		if (settingCurrent >= 5) {
			settingCurrent = 0;
		}
	}
	else { // Show current
		displayChar((int) currentNumber / 1000);
		drawDot(15, 22);
		xPos += 5;
		
		displayChar((int) (currentNumber / 100) % 10);
		displayChar((int) (currentNumber / 10) % 10);
		displayChar((int) currentNumber % 10);
		displayChar(FONT_A);
	}
	
	if (selectedOption == CHANGE_CURRENT) {
		xPos += 5;
		displayChar(FONT_ARROW);
	}
	
	// Current limited, display it
	if (currentNumber >= 5) {
		int currentDifference = currentNumber - currentLimitCalc;
		if (currentDifference >= -2) {
			xPos += 5;
			displayChar(FONT_CURRENT_LIMIT);
		}
	}
	
	// Watts
	xPos = 0;
	yPos = 44;
	
	uint16_t wattsNumber = ((uint32_t) voltageNumber * (uint32_t)currentNumber) / (uint32_t) 1000;
	if (wattsNumber >= 1000) {
		displayChar((int) wattsNumber / 1000);
		displayChar((int) (wattsNumber / 100) % 10);
		drawDot(30, 44);
		xPos += 5;
		
		displayChar((int) (wattsNumber / 10) % 10);
		displayChar((int) wattsNumber % 10);
		displayChar(FONT_W);
	}
	else {
		displayChar((int) (wattsNumber / 100) % 10);
		drawDot(15, 44);
		xPos += 5;
		
		displayChar((int) (wattsNumber / 10) % 10);
		displayChar((int) wattsNumber % 10);
		displayChar(0);
		displayChar(FONT_W);
	}
	
	Adafruit_SSD1306_display();
}

// ADC Read
int adc_read(uint8_t pin) {
	uint8_t low, high;
	
	ADMUX = (1<<REFS1) | (1<<REFS0) | (pin & 0x3f); // 1.1V voltage reference
	
	set_sleep_mode(SLEEP_MODE_ADC); // Set sleep mode ADC
	sleep_mode(); // System sleeps here
	
	// Read ADCL first
	low = ADCL;
	high = ADCH;
	
	// combine the two bytes
	return (high << 8) | low;
}

// Setup
void setup(void) {
	// Set outputs	
	DDRD = (1<<DPOT_CS) | (1<<DPOT_CLK) | (1<<LED_ENABLE) | (1<<EN_20_OHM) | (1<<EN_10_OHM) | (1<<EN_4_7_OHM) | (1<<EN_0_OHM);
	DDRC = (1<<DPOT_DATA);
	
	// Set pin states
	PORTD &= ~((1<<LED_ENABLE) | (1<<EN_10_OHM) | (1<<EN_4_7_OHM) | (1<<EN_0_OHM));
	PORTD |= (1<<EN_20_OHM);
	
	// Pin interrupt setup for the rotary encoder & select button
	PCICR |= (1<<PCIE0);
	PCMSK0 |= (1<<PCINT0) | (1<<PCINT1) | (1<<PCINT2);
	
	// Pin interrupt setup for enable button
	PCICR |= (1<<PCIE2);
	PCMSK2 |= (1<<PCINT23);
	
	// Turn on Timer1 interrupt on OCR1A compare
	TIMSK1 |= (1<<OCIE1A);
	TCCR1B |= (1<<WGM12);
	
	
	// Init I2C
	SoftI2cMasterInit(DAC_SCL, DAC_SDA);
	
	// Set VRef on DAC used for current limiting
	soft_i2c_write_byte(DAC_ADDR, WRITE_DAC_CONFIG_ENABLE_VREF, DAC_SCL, DAC_SDA);
	
	
	// Turn on ADC, ADC interrupt and set ADC prescale factor to 8
	// 12 MHz / 64 = 187 KHz, inside the desired 50-200 KHz range.
	ADCSRA = (1<<ADEN) | (1<<ADIE) | (1<<ADPS2) | (1<<ADPS1);
	
	// Setup screen
	Adafruit_SSD1306_begin(SSD1306_SWITCHCAPVCC);
	Adafruit_SSD1306_clearDisplay();
	
	// Turn on interrupt
	sei();
}
